import unittest

from mypyc.ir.rtypes import (
    RTuple, object_rprimitive, int_rprimitive, bool_rprimitive, list_rprimitive,
    RInstance, RUnion,
)
from mypyc.ir.class_ir import ClassIR


class TestTupleNames(unittest.TestCase):
    def setUp(self) -> None:
        self.inst_a = RInstance(ClassIR('A', '__main__'))
        self.inst_b = RInstance(ClassIR('B', '__main__'))

    def test_names(self) -> None:
        assert RTuple([int_rprimitive, int_rprimitive]).unique_id == "T2II"
        assert RTuple([list_rprimitive, object_rprimitive, self.inst_a]).unique_id == "T3OOO"
        assert RTuple([list_rprimitive, object_rprimitive, self.inst_b]).unique_id == "T3OOO"
        assert RTuple([]).unique_id == "T0"
        assert RTuple([RTuple([]),
                       RTuple([int_rprimitive, int_rprimitive])]).unique_id == "T2T0T2II"
        assert RTuple([bool_rprimitive,
                       RUnion([bool_rprimitive, int_rprimitive])]).unique_id == "T2CO"



"""Unit tests for file system cache."""

import os
import shutil
import sys
import tempfile
import unittest
from typing import Optional

from mypy.fscache import FileSystemCache


class TestFileSystemCache(unittest.TestCase):
    def setUp(self) -> None:
        self.tempdir = tempfile.mkdtemp()
        self.oldcwd = os.getcwd()
        os.chdir(self.tempdir)
        self.fscache = FileSystemCache()

    def tearDown(self) -> None:
        os.chdir(self.oldcwd)
        shutil.rmtree(self.tempdir)

    def test_isfile_case_1(self) -> None:
        self.make_file('bar.py')
        self.make_file('pkg/sub_package/__init__.py')
        self.make_file('pkg/sub_package/foo.py')
        # Run twice to test both cached and non-cached code paths.
        for i in range(2):
            assert self.isfile_case('bar.py')
            assert self.isfile_case('pkg/sub_package/__init__.py')
            assert self.isfile_case('pkg/sub_package/foo.py')
            assert not self.isfile_case('non_existent.py')
            assert not self.isfile_case('pkg/non_existent.py')
            assert not self.isfile_case('pkg/')
            assert not self.isfile_case('bar.py/')
        for i in range(2):
            assert not self.isfile_case('Bar.py')
            assert not self.isfile_case('pkg/sub_package/__init__.PY')
            assert not self.isfile_case('pkg/Sub_Package/foo.py')
            assert not self.isfile_case('Pkg/sub_package/foo.py')

    def test_isfile_case_2(self) -> None:
        self.make_file('bar.py')
        self.make_file('pkg/sub_package/__init__.py')
        self.make_file('pkg/sub_package/foo.py')
        # Run twice to test both cached and non-cached code paths.
        # This reverses the order of checks from test_isfile_case_1.
        for i in range(2):
            assert not self.isfile_case('Bar.py')
            assert not self.isfile_case('pkg/sub_package/__init__.PY')
            assert not self.isfile_case('pkg/Sub_Package/foo.py')
            assert not self.isfile_case('Pkg/sub_package/foo.py')
        for i in range(2):
            assert self.isfile_case('bar.py')
            assert self.isfile_case('pkg/sub_package/__init__.py')
            assert self.isfile_case('pkg/sub_package/foo.py')
            assert not self.isfile_case('non_existent.py')
            assert not self.isfile_case('pkg/non_existent.py')

    def test_isfile_case_3(self) -> None:
        self.make_file('bar.py')
        self.make_file('pkg/sub_package/__init__.py')
        self.make_file('pkg/sub_package/foo.py')
        # Run twice to test both cached and non-cached code paths.
        for i in range(2):
            assert self.isfile_case('bar.py')
            assert not self.isfile_case('non_existent.py')
            assert not self.isfile_case('pkg/non_existent.py')
            assert not self.isfile_case('Bar.py')
            assert not self.isfile_case('pkg/sub_package/__init__.PY')
            assert not self.isfile_case('pkg/Sub_Package/foo.py')
            assert not self.isfile_case('Pkg/sub_package/foo.py')
            assert self.isfile_case('pkg/sub_package/__init__.py')
            assert self.isfile_case('pkg/sub_package/foo.py')

    def test_isfile_case_other_directory(self) -> None:
        self.make_file('bar.py')
        with tempfile.TemporaryDirectory() as other:
            self.make_file('other_dir.py', base=other)
            self.make_file('pkg/other_dir.py', base=other)
            assert self.isfile_case(os.path.join(other, 'other_dir.py'))
            assert not self.isfile_case(os.path.join(other, 'Other_Dir.py'))
            assert not self.isfile_case(os.path.join(other, 'bar.py'))
            if sys.platform in ('win32', 'darwin'):
                # We only check case for directories under our prefix, and since
                # this path is not under the prefix, case difference is fine.
                assert self.isfile_case(os.path.join(other, 'PKG/other_dir.py'))

    def make_file(self, path: str, base: Optional[str] = None) -> None:
        if base is None:
            base = self.tempdir
        fullpath = os.path.join(base, path)
        os.makedirs(os.path.dirname(fullpath), exist_ok=True)
        if not path.endswith('/'):
            with open(fullpath, 'w') as f:
                f.write('# test file')

    def isfile_case(self, path: str) -> bool:
        return self.fscache.isfile_case(os.path.join(self.tempdir, path), self.tempdir)



"""Basic introspection of modules."""

from typing import List, Optional, Union
from types import ModuleType
from multiprocessing import Process, Queue
import importlib
import inspect
import os
import pkgutil
import queue
import sys


class ModuleProperties:
    def __init__(self,
                 name: str,
                 file: Optional[str],
                 path: Optional[List[str]],
                 all: Optional[List[str]],
                 is_c_module: bool,
                 subpackages: List[str]) -> None:
        self.name = name  # __name__ attribute
        self.file = file  # __file__ attribute
        self.path = path  # __path__ attribute
        self.all = all  # __all__ attribute
        self.is_c_module = is_c_module
        self.subpackages = subpackages


def is_c_module(module: ModuleType) -> bool:
    if module.__dict__.get('__file__') is None:
        # Could be a namespace package. These must be handled through
        # introspection, since there is no source file.
        return True
    return os.path.splitext(module.__dict__['__file__'])[-1] in ['.so', '.pyd']


class InspectError(Exception):
    pass


def get_package_properties(package_id: str) -> ModuleProperties:
    """Use runtime introspection to get information about a module/package."""
    try:
        package = importlib.import_module(package_id)
    except BaseException as e:
        raise InspectError(str(e)) from e
    name = getattr(package, "__name__", package_id)
    file = getattr(package, "__file__", None)
    path: Optional[List[str]] = getattr(package, "__path__", None)
    if not isinstance(path, list):
        path = None
    pkg_all = getattr(package, '__all__', None)
    if pkg_all is not None:
        try:
            pkg_all = list(pkg_all)
        except Exception:
            pkg_all = None
    is_c = is_c_module(package)

    if path is None:
        # Object has no path; this means it's either a module inside a package
        # (and thus no sub-packages), or it could be a C extension package.
        if is_c:
            # This is a C extension module, now get the list of all sub-packages
            # using the inspect module
            subpackages = [package.__name__ + "." + name
                           for name, val in inspect.getmembers(package)
                           if inspect.ismodule(val)
                           and val.__name__ == package.__name__ + "." + name]
        else:
            # It's a module inside a package.  There's nothing else to walk/yield.
            subpackages = []
    else:
        all_packages = pkgutil.walk_packages(path, prefix=package.__name__ + ".",
                                             onerror=lambda r: None)
        subpackages = [qualified_name for importer, qualified_name, ispkg in all_packages]
    return ModuleProperties(name=name,
                            file=file,
                            path=path,
                            all=pkg_all,
                            is_c_module=is_c,
                            subpackages=subpackages)


def worker(tasks: 'Queue[str]',
           results: 'Queue[Union[str, ModuleProperties]]',
           sys_path: List[str]) -> None:
    """The main loop of a worker introspection process."""
    sys.path = sys_path
    while True:
        mod = tasks.get()
        try:
            prop = get_package_properties(mod)
        except InspectError as e:
            results.put(str(e))
            continue
        results.put(prop)


class ModuleInspect:
    """Perform runtime introspection of modules in a separate process.

    Reuse the process for multiple modules for efficiency. However, if there is an
    error, retry using a fresh process to avoid cross-contamination of state between
    modules.

    We use a separate process to isolate us from many side effects. For example, the
    import of a module may kill the current process, and we want to recover from that.

    Always use in a with statement for proper clean-up:

      with ModuleInspect() as m:
          p = m.get_package_properties('urllib.parse')
    """

    def __init__(self) -> None:
        self._start()

    def _start(self) -> None:
        self.tasks: Queue[str] = Queue()
        self.results: Queue[Union[ModuleProperties, str]] = Queue()
        self.proc = Process(target=worker, args=(self.tasks, self.results, sys.path))
        self.proc.start()
        self.counter = 0  # Number of successful roundtrips

    def close(self) -> None:
        """Free any resources used."""
        self.proc.terminate()

    def get_package_properties(self, package_id: str) -> ModuleProperties:
        """Return some properties of a module/package using runtime introspection.

        Raise InspectError if the target couldn't be imported.
        """
        self.tasks.put(package_id)
        res = self._get_from_queue()
        if res is None:
            # The process died; recover and report error.
            self._start()
            raise InspectError('Process died when importing %r' % package_id)
        if isinstance(res, str):
            # Error importing module
            if self.counter > 0:
                # Also try with a fresh process. Maybe one of the previous imports has
                # corrupted some global state.
                self.close()
                self._start()
                return self.get_package_properties(package_id)
            raise InspectError(res)
        self.counter += 1
        return res

    def _get_from_queue(self) -> Union[ModuleProperties, str, None]:
        """Get value from the queue.

        Return the value read from the queue, or None if the process unexpectedly died.
        """
        max_iter = 100
        n = 0
        while True:
            if n == max_iter:
                raise RuntimeError('Timeout waiting for subprocess')
            try:
                return self.results.get(timeout=0.05)
            except queue.Empty:
                if not self.proc.is_alive():
                    return None
            n += 1

    def __enter__(self) -> 'ModuleInspect':
        return self

    def __exit__(self, *args: object) -> None:
        self.close()
